import java.util.Vector;
import org.w3c.dom.*;

/**
 * Knoten eines Baumes.
 * @author David Kaltschmidt & Benjamin Schrter
 */
public class Node {
	
	/** Array der Navigationsschlssel im Baum */
	int[] keys;
	/** Array der Schlssel (Bltter) */
	int[] leaves;
	/** Array der Kindknoten */
	Node[] nodes;
	/** Referenz auf den Vaterknoten  */
	Node father;
	/** Anzahl der Kinder (Knoten oder Bltter) */
	int size;

	/**
	 * Konstruktor des Knotens
	 * 
	 * @param type potenzielle Gre des Knotens
	 * @param f Vater des Knotens (bei Wurzel: f = null)
	 */
	public Node(int type, Node f) {
		size = 0;
		keys = new int[type+1];
		nodes = new Node[type+2];
		leaves = new int[type+2];
		father = f;
	}

	/**
	 * Sagt aus, ob der aktuelle Knoten Bltter hat.
	 * Ein Knoten hat Bltter, wenn er keine inneren Knoten
	 * als Kinder hat.
	 * 
	 * @return Boolean, ob der aktuelle Knoten Bltter hat.
	 */	
	public boolean hasLeaves() {
		return (nodes[0] == null);
	}
	
	/**
	 * Sagt aus, ob der aktuelle Knoten innere Knoten als Kinder hat.
	 * Ein Knoten hat innere Knoten als Kinder, wenn er keine Bltter hat.
	 * 
	 * @return Boolean, ob der aktuelle Knoten innere Knoten als Kinder hat.
	 */	
	public boolean hasNodes() {
		return (!hasLeaves());
	}

	/**
	 * Fllt den Vektor mit allen Schlsseln des Baumes.
	 * Verzweigt sich fr jeden Kindknoten.
	 * 
	 * @param v Vektor, der mit den Schlsseln gefllt werden soll
	 */
	public void traverse(Vector v) {
		for(int i = 0; i < size; i++)
			if(hasLeaves())
				v.add(new Integer(leaves[i]));
			else 
				nodes[i].traverse(v);
	}
	
	/**
	 * Sagt aus, ob der Schlssel k im Teilbaum unter dem aktuellen Knoten
	 * enthalten ist.
	 * Der Weg wird durch Vergleich mit den Navigationsschlsseln
	 * gefunden (Branching) und dann der entsprechende Unterbaum 
	 * betrachtet. 
	 * Wenn man schon bei den Blttern angekommen ist, wird verglichen,
	 * ob der Schlssel dem angepeilten Blatt entspricht.
	 * 
	 * @param k Schlssel, der gesucht werden soll.
	 * @return Boolean, ob der Schlssel enthalten ist.
	 */
	public boolean contains(int k) {
		int i = 0;
		while(i < size - 1 && k >= keys[i])
			i++;
		if(hasLeaves())
			return (k == leaves[i]);
		else
			return nodes[i].contains(k);	
	}

	/**
	 * Fgt die Navigationsschlssel sowie die Bltter in den Vektor ein.
	 * Fr jeden Knoten wird dabei ein (bzw. zwei) Einzelstring(s) erzeugt.
	 * 
	 * @param v Vektor, welcher die Einzelstrings aufnimmt.
	 */
	public void print(Vector v) {
		String s = new String("");
		s += "[";
		for(int i = 0; i < size - 1; i++) {
			s += keys[i] + " ";
		}
		s += "]\n";
		v.add(s);
		if(hasLeaves()) {
			String t = new String("");
			for(int i = 0; i < size; i++)
				t += "(" + leaves[i] + ")";
			v.add(t);
		}
			
		else
			for(int i = 0; i < size; i++) {
				nodes[i].print(v);
			}
	}

	/**
	 * Fgt einen Schlssel in den Teilbaum unter dem aktuellen Knoten ein.
	 * Das Einfgen luft in folgenden Schritten ab: <BR>
	 * - Einfgeposition i ber die Navigationsschlssel bestimmen <BR>
	 * - Wenn man noch nicht bei den Blttern ist, dann betrachte i-ten Unterbaum <BR>
	 * - Sonst macht man je nach Einfgepostion in den Navigationsschlssel
	 * sowie in den Blttern Platz (zwischen der i-ten Position),
	 * fgt den Schlssel ein, erhht die Gre um 1 und teilt den Knoten, 
	 * wenn er zu voll ist
	 * 
	 * @param t gesamter Baum
	 * @param key einzufgender Schlssel
	 * @see split(int, MyTree)
	 */
	public void add(int k, MyTree t) {
		//Einfuegeposition bestimmen
		int i = 0;
		while(i < size - 1 && k >= keys[i])
			i++;
		if(hasLeaves()) {
			//Schon vorhanden
			if(k == leaves[i]) {
				System.out.println("Schlssel schon vorhanden.");
				return;
			}
			//In den Keys Platz machen
			if(size > 0 && k < keys[i]) {
				for(int j = size - 1; j > i; j--)
					keys[j] = keys[j-1];
			}
			
			//Neuen Key schreiben
			if(k < leaves[i])
				keys[i] = leaves[i];
			else if(keys[0] == 0 && keys[1] == 0) //Leerer Node
				keys[0] = k;
			else
				keys[i] = k;

			
			//In den leaves Platz machen
			if(k > leaves[i] && !(leaves[i] == 0 && leaves[i+1] == 0))
				i++;
			if(k < leaves[i] && !(leaves[i] == 0 && leaves[i+1] == 0))
				for(int j = size; j > i; j--) {
					leaves[j] = leaves[j-1];
				}
			leaves[i] = k;
			size++;
			t.size++;
			if(size > t.type+1) 
				split(k, t);
			else if(t.height == 0)
				t.height++;
		}
		else
			nodes[i].add(k, t);
	}

	/**
	 * Teilt den aktuellen Knoten.
	 * Wenn aktuelle Knoten die Wurzel ist, wird eine neue erzeugt
	 * und die alte Wurzel unten angehangen.
	 * Ansonsten wird einfach ein neuer Knoten erzeugt, welcher die 
	 * zweite Hlfte der Navigationsschlssel/Kindknoten/Bltter bekommt.
	 * Anschlieend werden noch die Gren der Knoten korrigiert, sowie
	 * der neue Knoten im Vater verankert. Dazu wird auch wieder die
	 * Einfgeposition anhand des Schlssels bestimmt und gegebenfalls noch
	 * Platz gemacht (vgl. add), bevor der neue Navigationsschlssel (fatherKey)
	 * sowie der neue Knoten im Vater gesetzt werden.
	 * Wenn der Vater dann auch berfllt ist, soo wird er auch geteilt.
	 * 
	 * @param k letzter einzufgender Schlssel
	 * @param t gesamter Baum
	 */
	public void split(int k, MyTree t) {
		
		System.out.println("Splitting Node: " + keys[0]);
		
		//Root spezial 
		if(this.father == null) {
			System.out.println("Creating new ROOT");
			Node r = this;
			t.root = null;
			Node newRoot = new Node(t.type, null);
			r.father = newRoot;
			newRoot.nodes[0] = r;
			newRoot.size++;
			t.root = newRoot;
			t.height++;
		}


		//Erzeuge neuen Schwesterknoten z
		Node z = new Node(t.type, this.father);

		//z bekommt die Haelfte der hohen Keys
		for(int j = 0; j < t.type - 1; j++)
			z.keys[j] = keys[t.type + j];
		
		//sowie die hohen Knoten/Blaetter
		for(int j = 0; j < t.type; j++) {
			if(hasLeaves())
				z.leaves[j] = leaves[t.type + j];
			else {
				z.nodes[j] = nodes[t.type + j];
				z.nodes[j].father = z;
			}
		}
		
		//Korrektur der Groessen
		size = t.type;
		z.size = t.type;
		
		//Verankerung im Vaterknoten
		int fatherKey = keys[(int)t.type / 2];
		//Einfgeposition bestimmen
		int i = 0;
		while(i < father.size - 1 && fatherKey > father.keys[i])
			i++;
		//Platz machen in Keys und Nodes
		if(father.size > 0 && k < father.keys[i]) {
			for(int j = father.size - 1; j > i; j--) {
				father.keys[j] = father.keys[j-1];
				father.nodes[j+1] = father.nodes[j];
			}
		}
		System.out.println("Inserting Fatherkey " + fatherKey + " at Position: " + i + " in Father " + father.keys[0]);

		//Setzen von Key und neuem Node
		father.keys[i] = fatherKey;
		father.nodes[i+1] = z;
		father.size++;
		
		//Wenn Vaterknoten auch voll, dann split
		if(father != null && father.size > t.type + 1)
			father.split(k, t);
	}
	
	
	/**
	 * Erzeugt aus den Baumknoten XML-Knoten.
	 * 
	 * @param document XML-Dokument, wird zum erzeugen der Elemente bentigt
	 * @param element Vaterelement (umgewandelter Knoten), an den die erzeugten Elemente angehangen werden
	 * */
	public void toXml(Document document, Element element) {
		String s  = "keys_";
		for(int i = 0; i < size-1; i++)
			s += keys[i] +"_";
		
		Element e = (Element) document.createElement(s);
		
		if (element == null)
			document.appendChild(e);
		else
			element.appendChild(e);
		
		for(int i = 0; i < size; i++)
			if(hasLeaves()) {
				Element le = document.createElement("leaf"); 
				Text lt = document.createTextNode( Integer.toString(leaves[i]) );
				le.appendChild(lt);
				e.appendChild(le);
			}
			else {
				nodes[i].toXml(document, e); 
			}
	}
	
}
